[{% for msg in messages %}
  {
    {% for key, value in msg.items() -%}
      "{{ key }}": {{ value | tojson }}{% if not loop.last %},{% endif %}
    {% endfor %}
  }{% if not loop.last %},{% endif %}
{% endfor %}]